(* Imports. *)

Require Import LibTactics.
Require Coq.omega.Omega.
Require Import Coq.Program.Equality.
Require Import Coq.Logic.ProofIrrelevance.
Require Import ArithMisc.
Unset Implicit Arguments.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Tactics. *)

Ltac myOmega f :=
  assert f as evidence; [ omega | rewrite evidence; clear evidence; auto ].

Ltac myProofIrrelevance p q :=
  assert (p = q) as fact; [ apply proof_irrelevance | repeat (rewrite fact); clear fact ].

Ltac myEqualUpToPI :=
  repeat progress f_equal; try apply proof_irrelevance.

Tactic Notation "myPunchPI" constr(x) :=
  let t := type of x in
  let q := fresh in
  evar (q : t); rewrite (proof_irrelevance _ x q); unfold q; clear q.

Tactic Notation "myPunchPI" constr(x) "in" hyp(H) :=
  let t := type of x in
  let q := fresh in
  evar (q : t); rewrite (proof_irrelevance _ x q) in H; unfold q in H; clear q.


(* ----------------------------------------------------------------------------------------------------------------- *)

(* Types. *)

(* In Nakano's presentation, there is an explicit \mu constructor, together
   with explicit subtyping rules for folding and unfolding recursive
   types. Here, instead, recursive types are represented as potentially
   infinite trees. This seems more appealing. *)

CoInductive ty : Type :=
| TyLeaf : ty
| TyArrow : ty -> ty -> ty
| TyBullet : ty -> ty.

(* A decomposition function and lemma, in the style of Bertot and Castran. *)

Definition ty_decomp (t : ty) : ty :=
  match t with
  | TyLeaf => TyLeaf
  | TyArrow u v => TyArrow u v
  | TyBullet t => TyBullet t
  end.

Lemma ty_decompose:
  forall t, t = ty_decomp t.
Proof.
  intro t; case t; eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Type equality. *)

CoInductive tyeq : ty -> ty -> Prop :=
| EqLeaf : tyeq TyLeaf TyLeaf
| EqArrow : forall u1 u2 v1 v2, tyeq u1 v1 -> tyeq u2 v2 -> tyeq (TyArrow u1 u2) (TyArrow v1 v2)
| EqBullet : forall t u, tyeq t u -> tyeq (TyBullet t) (TyBullet u).

Lemma tyeq_reflexive:
  forall t, tyeq t t.
Proof.
  cofix.
  intro t; case t; econstructor; eauto.
Qed.

Hint Resolve tyeq_reflexive.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The predicate transformer [everywhere] is used to require a property of a
   type to hold everywhere in a type, that is, at every sub-term. *)

CoInductive everywhere (P : ty -> Prop) : ty -> Prop :=
| EverywhereLeaf : P TyLeaf -> everywhere P TyLeaf
| EverywhereArrow : forall u v, everywhere P u -> everywhere P v -> P (TyArrow u v) -> everywhere P (TyArrow u v)
| EverywhereBullet : forall t, everywhere P t -> P (TyBullet t) -> everywhere P (TyBullet t).

Hint Constructors everywhere.

Lemma everywhere_root:
  forall P t, everywhere P t -> P t.
Proof.
  intros ? ? evpt. case evpt; eauto.
Qed.

Lemma everywhere_hereditary_bullet:
  forall P t, everywhere P (TyBullet t) -> everywhere P t.
Proof.
  intros ? ? h. dependent destruction h; eauto.
Qed.

Lemma everywhere_hereditary_domain:
  forall P u v, everywhere P (TyArrow u v) -> everywhere P u.
Proof.
  intros ? ? ? h. dependent destruction h; eauto.
Qed.

Lemma everywhere_hereditary_codomain:
  forall P u v, everywhere P (TyArrow u v) -> everywhere P v.
Proof.
  intros ? ? ? h. dependent destruction h; eauto.
Qed.

Hint Resolve everywhere_hereditary_bullet.
Hint Resolve everywhere_hereditary_domain.
Hint Resolve everywhere_hereditary_codomain.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Finiteness. *)

(* A type is tail-finite if and only if its rightmost branch is finite,
   that is, if and only if its rightmost branch reaches a leaf. *)

Inductive tf : ty -> Prop :=
| TfLeaf : tf TyLeaf
| TfArrow : forall u v, tf v -> tf (TyArrow u v)
| TfBullet : forall t, tf t -> tf (TyBullet t).

Hint Constructors tf.

(* Induction over a tail-finiteness property often requires a strengthened
   induction scheme. *)

Scheme tf_prop := Induction for tf Sort Prop.

(* Tail-finiteness is hereditary. *)

(* The proof of the following lemma has to be very carefully formulated, so
   that the proof term can be considered by Coq as producing a subterm of the
   argument [tft]. This technique (recursion over an ad hoc predicate) is
   borrowed from Bertot. *)

Lemma tf_hereditary_bullet:
  forall t, tf t -> forall u, t = TyBullet u -> tf u.
Proof.
  intros ? tft.
  case tft.
  intros; discriminate.
  intros; discriminate.
  intros v tfv u bvbu. injection bvbu. intro vu. rewrite vu in tfv. assumption.
Defined.

(* A type is finite if and only if it is tail-finite and its immediate
   sub-terms are finite. In other words, a type is finite if and only
   if each of its sub-terms is tail-finite. *)

Definition finite (t : ty) :=
  everywhere tf t.

Hint Unfold finite.

(* Finiteness implies tail-finiteness. *)

Lemma finite_tf:
  forall t, finite t -> tf t.
Proof.
  apply everywhere_root.
Qed.

Hint Resolve finite_tf.

(* Finiteness is hereditary. *)

Lemma finite_hereditary_bullet:
  forall t, finite t -> forall u, t = TyBullet u -> finite u.
Proof.
  intros. subst. eauto.
Qed.

Lemma finite_hereditary_codomain:
  forall t, finite t -> forall u v, t = TyArrow u v -> finite v.
Proof.
  intros. subst. eauto.
Qed.

Lemma finite_hereditary_domain:
  forall t, finite t -> forall u v, t = TyArrow u v -> finite u.
Proof.
  intros. subst. eauto.
Qed.

(* Finiteness is preserved by constructors. *)

Lemma finite_arrow:
  forall u v, finite u -> finite v -> finite (TyArrow u v).
Proof.
  eauto.
Qed.

Lemma finite_bullet:
  forall t, finite t -> finite (TyBullet t).
Proof.
  eauto.
Qed.

Hint Resolve finite_arrow.
Hint Resolve finite_bullet.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* If a type is tail-finite, then it is possible to count how many bullets
   appear at its root. This is done by the function [head]. This function is
   defined by structural induction over a tail-finiteness property. *)

Fixpoint head (t : ty) (ft : finite t) (tft : tf t) { struct tft } : nat :=
  match t as b return t = b -> nat with
  | TyLeaf => fun _ => 0
  | TyArrow u v => fun _ => 0
  | TyBullet u => fun (heq : t = TyBullet u) =>
      S (head u (finite_hereditary_bullet t ft u heq) (tf_hereditary_bullet t tft u heq))
  end (refl_equal t).

(* We now prove that [head] satisfies the expected equations. *)

Lemma head_leaf:
  forall ft tft, head TyLeaf ft tft = 0.
Proof.
  intros.
  dependent destruction ft.
  dependent destruction tft.
  simpl.
  auto.
Qed.

Lemma head_arrow:
  forall u v ft tft, head (TyArrow u v) ft tft = 0.
Proof.
  intros.
  dependent destruction ft.
  dependent destruction tft.
  simpl.
  auto.
Qed.

Lemma head_bullet:
  forall u ft tft,
    let fu := finite_hereditary_bullet _ ft u refl in
    let tfu := tf_hereditary_bullet _ tft u refl in
    head (TyBullet u) ft tft = S (head u fu tfu).
Proof.
  intros.
  dependent destruction ft.
  dependent destruction tft.
  reflexivity.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* If a type is tail-finite, then it is possible to skip the bullets that
   appear at its root, and produce the type that follows. This is done by the
   function [tail]. Like [head], this function is defined by structural
   induction over a tail-finiteness property. *)

(* [tail] is somewhat more complex than [head], because it returns not only
   a type [t], but also a guarantee that [t] does not begin with a bullet
   and that [t] is finite. *)

(* The predicate [nab] is short for ``not-a-bullet''. *)

Inductive nab : ty -> Prop :=
| NabLeaf: nab TyLeaf
| NabArrow: forall u v, nab (TyArrow u v).

Hint Constructors nab.

(* This type is the codomain of [tail]. *)

Definition nabft := { t : ty | nab t /\ finite t }.

(* The TyLeaf and TyArrow must be suitably wrapped. *)

Definition nabft_leaf : nabft.
Proof.
  econstructor. split. apply NabLeaf. eauto.
Defined.

Definition nabft_arrow : forall t u v, t = TyArrow u v -> finite t -> nabft.
Proof.
  econstructor. split. apply (NabArrow u v). subst. eauto.
Defined.

(* Here comes the definition of [tail]. *)

Fixpoint tail (t : ty) (ft : finite t) (tft : tf t) { struct tft } : nabft :=
  match t as b return t = b -> nabft with
  | TyLeaf => fun _ =>
      (* produce a leaf *)
      nabft_leaf
  | TyArrow u v => fun heq =>
      (* produce an arrow *)
      nabft_arrow t u v heq ft
  | TyBullet u => fun (heq : t = TyBullet u) =>
      (* skip this bullet and go down *)
      tail u
        (finite_hereditary_bullet t ft u heq)
        (tf_hereditary_bullet t tft u heq)
  end (refl_equal t).

(* We now prove that [tail] satisfies the expected equations. *)

Lemma tail_leaf:
  forall ft tft, tail TyLeaf ft tft = nabft_leaf.
Proof.
  intros. dependent destruction tft. eauto.
Qed.

Lemma tail_arrow:
  forall u v ft tft, tail (TyArrow u v) ft tft = nabft_arrow (TyArrow u v) u v refl ft.
Proof.
  intros. dependent destruction tft. eauto.
Qed.

Lemma tail_bullet:
  forall t ft tft,
  tail (TyBullet t) ft tft =
  tail t
    (finite_hereditary_bullet _ ft _ refl)
    (finite_tf _ (finite_hereditary_bullet _ ft _ refl)).
Proof.
  intros. dependent destruction tft. simpl.
  f_equal. apply proof_irrelevance.
Qed.

(* The tail of a finite type is not a bullet, and is a finite type. *)

Lemma nabft_tail:
  forall t ft tft,
  nab (proj1_sig (tail t ft tft)) /\ finite (proj1_sig (tail t ft tft)).
Proof.
  intros. apply proj2_sig.
Qed.

Lemma finite_tail:
  forall t ft tft,
  finite (proj1_sig (tail t ft tft)).
Proof.
  intros. eapply nabft_tail.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Skeletons. *)

(* Skeletons are types in canonical form, where all bullets are pushed down to
   the leaves. As a result, there is no analogue of the [TyBullet] constructor.
   Leaves carry a finite number of bullets. Arrows carry no bullets whatsoever. *)

CoInductive skeleton : Type :=
| SkLeaf : nat -> skeleton
| SkArrow : skeleton -> skeleton -> skeleton.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Skeleton equality. *)

CoInductive skeq : skeleton -> skeleton -> Prop :=
| SkEqLeaf : forall m, skeq (SkLeaf m) (SkLeaf m)
| SkEqArrow : forall t1 t2 u1 u2, skeq t1 t2 -> skeq u1 u2 -> skeq (SkArrow t1 u1) (SkArrow t2 u2).

Hint Constructors skeq.

Lemma skeq_reflexive:
  forall s, skeq s s.
Proof.
  cofix. intro s. case s; constructor; auto.
Qed.

Lemma skeq_symmetric:
  forall s t, skeq s t -> skeq t s.
Proof.
  cofix. intros ? ? st. dependent destruction st; constructor; auto.
Qed.

Lemma skeq_transitive:
  forall s t u, skeq s t -> skeq t u -> skeq s u.
Proof.
  cofix. intros ? ? ? st tu. dependent destruction st; dependent destruction tu; constructor; eauto.
Qed.

Hint Resolve skeq_reflexive.
Hint Immediate skeq_symmetric.
Hint Resolve skeq_transitive.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Decomposition lemmas. *)

Definition sk_decomp (t : skeleton) : skeleton :=
  match t with
  | SkLeaf n => SkLeaf n
  | SkArrow u v => SkArrow u v
  end.

Lemma sk_decompose:
  forall t, t = sk_decomp t.
Proof.
  intro t; case t; eauto.
Qed.

Lemma skeq_decompose:
  forall s t,
  skeq (sk_decomp s) (sk_decomp t) ->
  skeq s t.
Proof.
  intros.
  rewrite (sk_decompose s).
  rewrite (sk_decompose t).
  assumption.
Defined.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Shifting a skeleton amounts to prepending a number of bullets at the root
   and pushing them down to the leaves, so as to obtain another skeleton. *)

CoFixpoint shift (k : nat) (s : skeleton) :=
  match s with
  | SkLeaf m => SkLeaf (k + m)
  | SkArrow t u => SkArrow (shift k t) (shift k u)
  end.

Lemma shift_zero:
  forall s,
  skeq (shift 0 s) s.
Proof.
  cofix. intros [ n | s1 s2 ]; rewrite (sk_decompose (shift 0 _)); simpl; auto.
Qed.

Lemma shift_shift:
  forall j k s,
  skeq (shift j (shift k s)) (shift (j + k) s).
Proof.
  cofix.
  intros.
  apply skeq_decompose.
  dependent destruction s; simpl.
  repeat (rewrite arith_plus_associative). eauto.
  eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Mapping types to skeletons. *)

(* Unfortunately, not every type has a skeleton: the type Top, defined as (\mu
   X. TyBullet X), has none, because there is no way to skip the bullets and
   find some structure. Only finite types have skeletons. *)

(* The following term is used to mark the TyBullet branch as impossible in the
   definition of [sk]. *)

Lemma nab_bullet_absurd:
  forall t u,
  nab t -> t = TyBullet u ->
  forall A : Set, A.
Proof.
  intros ? ? nabt ? ?. subst.
  eapply (@absurd_set True).
  trivial.
  dependent destruction nabt.
Qed.

(* The function [sk] maps a finite type [t] to its skeleton. The accumulator
   [m] is a number of bullets by which the skeleton should be shifted. The
   term [ft] is a proof of finiteness. *) 

CoFixpoint sk (m : nat) (t : ty) (ft : finite t) : skeleton :=

  (* [t] is finite, hence [t] is tail-finite. *)

  let tft := finite_tf t ft in

  (* Skipping [k] bullets at the root, extract the tail, [tt]. *)

  let k  : nat   := head t ft tft in
  let tt : nabft := tail t ft tft in

  (* The term [tt] has type [nabft], i.e., it is a type, wrapped
     with a proof that it is finite and is not a bullet. In the
     following, [proj1_sig tt] stands for the naked term, while
     [proj2_sig tt] stands for the proof, which can be further
     decomposed using [proj1] and [proj2]. *)
  
  match proj1_sig tt as b return proj1_sig tt = b -> skeleton with
  | TyLeaf => fun _ =>
      (* produce a leaf *)
      SkLeaf (m + k)
  | TyArrow u v => fun heq =>
      (* produce an arrow and continue *)
      let ftt := proj2 (proj2_sig tt) in
      SkArrow
        (sk (m + k) u (finite_hereditary_domain _ ftt _ _ heq))
        (sk (m + k) v (finite_hereditary_codomain _ ftt _ _ heq))
  | TyBullet _ => fun heq =>
      (* impossible *)
      let nabtt := proj1 (proj2_sig tt) in
      nab_bullet_absurd _ _ nabtt heq _
  end (refl_equal _).

(* We now prove that [sk] satisfies the expected equations. *)

(* There is a lot of noise in some of these equations, due to the proof term
   that [sk] requires as an argument. The axiom of proof irrelevance must
   sometimes be used to equate some of these proof terms. *)

Lemma sk_leaf:
  forall m ft, sk m TyLeaf ft = SkLeaf m.
Proof.
  intros.
  rewrite (sk_decompose (sk _ _ _)). simpl.
  rewrite tail_leaf. simpl. (* miracle! *)
  rewrite head_leaf.
  eauto.
Qed.

Lemma sk_arrow:
  forall m u v ft,
  sk m (TyArrow u v) ft =
  SkArrow
    (sk m u (finite_hereditary_domain _ ft _ _ refl))
    (sk m v (finite_hereditary_codomain _ ft _ _ refl)).
Proof.
  intros.
  rewrite (sk_decompose (sk _ (TyArrow u v) ft)). simpl.
  rewrite tail_arrow. simpl.
  rewrite head_arrow.
  unfold eq_ind_r. simpl. (* more miracles! *)
  myEqualUpToPI; omega.
Qed.

Lemma sk_bullet:
  forall m t ft,
  sk m (TyBullet t) ft = sk (S m) t (finite_hereditary_bullet _ ft _ refl).
Proof.
  intros.
  rewrite (sk_decompose (sk _ (TyBullet t) ft)); simpl.
  rewrite (sk_decompose (sk _ t _)); simpl.
  rewrite tail_bullet.
  rewrite head_bullet.
  repeat (rewrite arith_plus_successor).
  (* At this point, the two expressions are equal up to proof-irrelevant
     terms, but [f_equal] does not work, because the two expressions are
     [match] constructs. We force an eta-expansion so as to allow the
     pmatch] constructs to disappear. *)
  generalize (tail t (finite_hereditary_bullet _ ft t refl) (finite_tf t (finite_hereditary_bullet _ ft t refl)));
  intros [ tt [ nabt ? ]]. simpl.
  dependent destruction tt; dependent destruction nabt; myEqualUpToPI.
Qed.

(* Prove that the accumulator serves to simulate a shift. *)

(* For modularity, one could wish to remove the accumulator and simply use
   [shift] in the definition of [sk], but that would unfortunately violate the
   guard condition. *)

Lemma sk_shift:
  forall t ft k m,
  skeq (sk (k + m) t ft) (shift k (sk m t ft)).
Proof.
  cofix.
  intros.
  apply skeq_decompose; simpl.
  generalize (tail t ft (finite_tf t ft));
  intros [ tt [ nabt ? ]]. simpl.
  rewrite <- arith_plus_associative.
  dependent destruction tt; simpl.
  (* Leaf *)
  constructor.
  (* Arrow *)
  constructor; eauto.
  (* Bullet *)
  dependent destruction nabt.
Qed.

Lemma sk_shift_zero:
  forall t ft k,
  skeq (sk k t ft) (shift k (sk 0 t ft)).
Proof.
  intros.
  replace (sk k t ft) with (sk (k + 0) t ft).
  apply sk_shift. rewrite arith_plus_zero. auto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Subtyping over skeletons. *)

(* Because skeletons have simple structure, the definition of subtyping over
   skeletons is quite straightforward. At leaves, we count the bullets. At
   arrows, we apply the standard contravariance-covariance rule, with a slight
   twist: it is permitted to prepend [k] bullets to the left-hand arrow before
   applying the standard rule. *)

(* The definition of subtyping has built-in shifts, so [sksub i t j u] means
   we are comparing [shift i t] with [shift j u]. *)

CoInductive sksub : nat -> skeleton -> nat -> skeleton -> Prop :=
| SkSubLeaf:
  forall i m j n,
    i + m <= j + n ->
    sksub i (SkLeaf m) j (SkLeaf n)
| SkSubArrow :
  forall k i t1 t2 j u1 u2,
    sksub j u1 (i + k) t1 ->
    sksub (i + k) t2 j u2 ->
    sksub i (SkArrow t1 t2) j (SkArrow u1 u2).

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Basic lemmas about subtyping over skeletons. *)

(* A decomposition lemma. *)

Lemma sksub_decompose:
  forall i s j t,
  sksub i (sk_decomp s) j (sk_decomp t) ->
  sksub i s j t.
Proof.
  intros.
  rewrite (sk_decompose s).
  rewrite (sk_decompose t).
  assumption.
Defined.

(* Subtyping is compatible with equality. *)

Lemma skeq_sksub_skeq:
  forall i j s t u v,
  skeq s t ->
  sksub i t j u ->
  skeq u v ->
  sksub i s j v.
Proof.
  cofix.
  intros ? ? ? ? ? ? st tu uv.
  dependent destruction st; dependent destruction tu; dependent destruction uv.
  constructor. omega.
  apply (SkSubArrow k); eauto 2. (* prevent eauto from searching too far and violating the guard condition! *)
Qed.

(* Subtyping is reflexive. *)

Lemma sksub_reflexive:
  forall i t, sksub i t i t.
Proof.
  cofix.
  intros.
  dependent destruction t.
  (* Leaf *)
  econstructor. eauto.
  (* Arrow *)
  eapply (SkSubArrow 0); rewrite arith_plus_zero; eauto.
Qed.

(* Subtyping is contravariant in its left index. *)

Lemma sksub_index_left:
  forall i1 i2 t j u,
  i1 <= i2 ->
  sksub i2 t j u ->
  sksub i1 t j u.
Proof.
  intros ? ? ? ? ? ? h.
  destruct h.
  (* Leaf *)
  constructor. omega.
  (* Arrow *)
  apply (SkSubArrow (i + k - i1)); rewrite arith_cancellation; try omega; auto.
Qed.

(* Only the difference between the two indices matters. *)

Lemma sksub_index_relative:
  forall i1 i2 t j1 j2 u,
  sksub i1 t j1 u ->
  i1 + j2 = j1 + i2 ->
  sksub i2 t j2 u.
Proof.
  cofix.
  intros ? ? ? ? ? ? h ?.
  destruct h.
  (* Leaf *)
  constructor. omega.
  (* Arrow *)
  apply (SkSubArrow k).
  eapply (sksub_index_relative _ _ _ _ _ _ h1); omega.
  eapply (sksub_index_relative _ _ _ _ _ _ h2); omega.
Qed.

Lemma sksub_index_both:
  forall i t j u,
  sksub i t j u ->
  forall n,
  sksub (i + n) t (j + n) u.
Proof.
  intros ? ? ? ? h n.
  eapply sksub_index_relative.
  eassumption.
  omega.
Qed.

(* Subtyping is covariant in its right index. *)

Lemma sksub_index_right:
  forall i t j1 j2 u,
  j1 <= j2 ->
  sksub i t j1 u ->
  sksub i t j2 u.
Proof.
  cofix.
  intros ? ? ? ? ? ? h.
  dependent destruction h.
  (* Leaf *)
  econstructor. omega.
  (* Arrow *)
  (* This is more subtle than one would think... and not symmetric
     with respect to the previous lemma, because the definition of
     sksub is asymmetric. *)
  rewrite <- (arith_cancellation j j2).
  apply (SkSubArrow (k + (j2 - j)));
  rewrite arith_plus_associative;
  auto using sksub_index_both.
  omega.
Qed.

(* Subtyping is transitive. *)

Lemma sksub_transitive:
  forall i j k t u v,
  sksub i t j u ->
  sksub j u k v ->
  sksub i t k v.
Proof.
  cofix.
  intros ? ? ? ? ? ? tu uv.
  dependent destruction tu; dependent destruction uv.
  (* Leaf *)
  constructor. omega.
  (* Arrow *)
  apply (SkSubArrow (k0 + k));
  rewrite arith_plus_associative;
  eapply sksub_transitive;
  eauto using sksub_index_both.
Qed.

(* As claimed informally above, the indices play the role of shifts. *)

Lemma sksub_shift:
  forall i1 i2 s j1 j2 t,
  sksub (i1 + i2) s (j1 + j2) t -> sksub i1 (shift i2 s) j1 (shift j2 t).
Proof.
  cofix.
  intros ? ? ? ? ? ? h.
  apply sksub_decompose; simpl.
  dependent destruction h.
  (* Leaf *)
   econstructor. omega.
  (* Arrow *)
  apply (SkSubArrow k);
  apply sksub_shift;
  rewrite arith_plus_commutative_23;
  assumption.
Qed.

Lemma sksub_plus:
  forall i1 i2 s j1 j2 t,
  sksub i1 (shift i2 s) j1 (shift j2 t) ->
  sksub (i1 + i2) s (j1 + j2) t.
Proof.
  cofix.
  intros ? ? ? ? ? ? h.
  rewrite (sk_decompose (shift i2 s)) in h.
  rewrite (sk_decompose (shift j2 t)) in h.
  simpl in h.
  dependent destruction s;
  dependent destruction t;
  dependent destruction h.
  (* Leaf *)
   econstructor. omega.
  (* Arrow *)
  apply (SkSubArrow k);
  rewrite arith_plus_commutative_23;
  apply sksub_plus;
  assumption.
Qed.

(* When comparing two skeletons obtained via [sk], it is equivalent to
   provide [j] as a parameter to [sk] or as a parameter to [sksub]. *)

Lemma sksub_plus_sk:
  forall t u ft fu i1 j1 i2 j2,
  sksub i1 (sk j1 t ft) i2 (sk j2 u fu) ->
  sksub (i1 + j1) (sk 0 t ft) (i2 + j2) (sk 0 u fu).
Proof.
  introv h.
  apply sksub_plus.
  eapply skeq_sksub_skeq.
  apply skeq_symmetric. apply sk_shift_zero. eassumption.
  apply sk_shift_zero.
Qed.

Lemma sksub_sk_plus:
  forall t u ft fu i1 i2 j1 j2,
  sksub (i1 + j1) (sk 0 t ft) (i2 + j2) (sk 0 u fu) ->
  sksub i1 (sk j1 t ft) i2 (sk j2 u fu).
Proof.
  intros ? ? ? ? ? ? ? ? h.
  eapply skeq_sksub_skeq.
  apply sk_shift_zero.
  apply sksub_shift. eassumption.
  apply skeq_symmetric.
  apply sk_shift_zero.
Qed.

(* A proof-irrelevance-related auxiliary lemma. *)

Lemma sksub_pi:
  forall i1 j1 t ft1 ft2 i2 j2 u fu1 fu2,
  sksub i1 (sk j1 t ft1) i2 (sk j2 u fu1) ->
  sksub i1 (sk j1 t ft2) i2 (sk j2 u fu2).
Proof.
  intros.
  myProofIrrelevance ft2 ft1.
  myProofIrrelevance fu2 fu1.
  assumption.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Subtyping over types. *)

(* Subtyping over types is defined in terms of subtyping over skeletons: two
   finite types are related if and only if their skeletons are related. *)

(* A drawback of this approach is that types that are not finite (that is,
   types that have Top as a sub-term) cannot participate in the subtying
   relation. This is hopefully acceptable, since these types are ruled out
   anyway in the normalization result. *)

Inductive sub : ty -> ty -> Prop :=
| Sub: forall t u ft fu, sksub 0 (sk 0 t ft) 0 (sk 0 u fu) -> sub t u.

Hint Constructors sub.

(* Subtyping is defined over finite types only. *)

Lemma sub_finite_left:
  forall t u, sub t u -> finite t.
Proof.
  introv tu. case tu. auto.
Qed.

Lemma sub_finite_right:
  forall t u, sub t u -> finite u.
Proof.
  introv tu. case tu. auto.
Qed.

Hint Resolve sub_finite_left sub_finite_right.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Flexible lemmas for moving from [sub] to [sksub], and back. *)

Lemma sub_sksub:
  forall t u,
  sub t u ->
  forall i j ft fu, sksub i (sk j t ft) i (sk j u fu).
Proof.
  intros ? ? h ? ? ? ?.
  destruct h.
  apply sksub_sk_plus.
  eapply sksub_index_relative.
  eapply sksub_pi. eassumption.
  omega.
Qed.

Lemma sksub_sub:
  forall t u i j ft fu,
  sksub i (sk j t ft) i (sk j u fu) ->
  sub t u.
Proof.
  intros ? ? ? ? ? ? h.
  apply (Sub t u ft fu).
  forwards: sksub_plus_sk. eexact h.
  eapply sksub_index_relative. eassumption. omega.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now check that all of the expected principles for building subtyping
   judgements are valid. *)

(* Subtyping is reflexive. *)

Lemma sub_reflexive:
  forall t, finite t -> sub t t.
Proof.
  intros t ft.
  apply (Sub t t ft ft).
  apply sksub_reflexive.
Qed.

Hint Resolve sub_reflexive.

(* Arrow is contravariant in its domain and covariant in its codomain. *)

Lemma sub_arrow:
  forall t1 t2 u1 u2, sub u1 t1 -> sub t2 u2 -> sub (TyArrow t1 t2) (TyArrow u1 u2).
Proof.
  intros ? ? ? ? u1t1 t2u2.
  dependent destruction u1t1; dependent destruction t2u2.
  assert (finite (TyArrow u t0)) as fut0. auto.
  assert (finite (TyArrow t u0)) as ftu0. auto.
  apply (Sub _ _ fut0 ftu0).
  rewrite (sk_arrow 0 u t0 fut0).
  rewrite (sk_arrow 0 t u0 ftu0).
  apply (SkSubArrow 0);
  eauto using sub_sksub.
Qed.

(* Bullet is covariant. *)

Lemma sub_bullet:
  forall u v, sub u v -> sub (TyBullet u) (TyBullet v).
Proof.
  intros ? ? uv.
  destruct uv.
  assert (finite (TyBullet t)) as ftt. eauto.
  assert (finite (TyBullet u)) as ftu. eauto.
  apply (Sub _ _ ftt ftu).
  rewrite (sk_bullet 0 t ftt).
  rewrite (sk_bullet 0 u ftu).
  eauto using sub_sksub.
Qed.

(* [TyBullet t] is weaker than [t]. *)

Lemma sub_bullet_intro:
  forall u v, sub u v -> sub u (TyBullet v).
Proof.
  intros ? ? tu.
  destruct tu as [ ? ? ? ? h ].
  assert (finite (TyBullet u)) as ftu. auto.
  apply (Sub _ _ ft ftu).
  rewrite sk_bullet.
  apply sksub_sk_plus.
  apply (sksub_index_right _ _ 0); simpl.
  omega.
  eauto using sksub_pi.
Qed.

(* [TyBullet] distributes over [TyArrow]. *)

Lemma sub_arrow_bullet_down:
  forall u v,
  finite u -> finite v ->
  sub (TyBullet (TyArrow u v)) (TyArrow (TyBullet u) (TyBullet v)).
Proof.
  intros u v fu fv.
  assert (finite (TyBullet (TyArrow u v))) as fba. eauto.
  assert (finite (TyArrow (TyBullet u) (TyBullet v))) as fab. eauto.
  apply (Sub _ _ fba fab).
  repeat (rewrite sk_bullet).
  repeat (rewrite sk_arrow).
  repeat (rewrite sk_bullet).
  apply (SkSubArrow 0);
  eauto using sub_sksub.
Qed.

Lemma sub_arrow_bullet_up:
  forall u v,
  finite u -> finite v ->
  sub (TyArrow (TyBullet u) (TyBullet v)) (TyBullet (TyArrow u v)).
Proof.
  intros u v fu fv.
  assert (finite (TyBullet (TyArrow u v))) as fba. eauto.
  assert (finite (TyArrow (TyBullet u) (TyBullet v))) as fab. eauto.
  apply (Sub _ _ fab fba).
  repeat (rewrite sk_bullet).
  repeat (rewrite sk_arrow).
  repeat (rewrite sk_bullet).
  apply (SkSubArrow 0);
  eauto using sub_sksub.
Qed.

(* Subtyping is transitive. *)

Lemma sub_transitive:
  forall t u v,
  sub t u ->
  sub u v ->
  sub t v.
Proof.
  intros t u v tu uv.
  dependent destruction tu; dependent destruction uv.
  apply (Sub _ _ ft fu0).
  eapply sksub_transitive.
  eassumption.
  eauto using sksub_pi.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The following is a digression, possibly not useful for the rest of the
   development. *)

(* A sanity check: we prove that the subtyping relation is not trivial. In
   particular, a ``bullet-leaf'' (a type that contains no arrows) can only
   be related to another ``bullet-leaf''. *)

CoInductive bleaf : ty -> Prop :=
| BLeafLeaf : bleaf TyLeaf
| BLeafBullet : forall t, bleaf t -> bleaf (TyBullet t).

Hint Constructors bleaf.

Lemma bleaf_hereditary:
  forall t, bleaf (TyBullet t) -> bleaf t.
Proof.
  intros ? h.
  dependent destruction h.
  auto.
Qed.

Hint Resolve bleaf_hereditary.

(* A (finite) type satisfies [bleaf] if and only if its tail is [TyLeaf]. *)

Lemma bleaf_tail:
  forall t, forall ft: finite t, forall tft: tf t, bleaf t -> proj1_sig (tail t ft tft) = TyLeaf.
Proof.
  intros ? ? tft.
  dependent induction tft using tf_prop; intro bt.
  (* Leaf *)
  auto.
  (* Arrow (impossible) *)
  dependent destruction bt.
  (* Bullet *)
  apply IHtft; auto.
Qed.

Lemma tail_bleaf:
  forall t, forall ft: finite t, forall tft: tf t, proj1_sig (tail t ft tft) = TyLeaf -> bleaf t.
Proof.
  intros ? ? tft.
  dependent induction tft using tf_prop; intro bt.
  (* Leaf *)
  auto.
  (* Arrow (impossible) *)
  rewrite tail_arrow in bt. simpl in bt. discriminate.
  (* Bullet *)
  constructor. rewrite tail_bullet in bt. eapply IHtft; auto.
  rewrite <- bt. myEqualUpToPI. intros. myEqualUpToPI. (* weird *)
Qed.

(* A (finite) type satisfies [bleaf] if and only if its skeleton is [SkLeaf _]. *)

Lemma bleaf_sk:
  forall t, forall ft: finite t, bleaf t -> forall i, sk i t ft = SkLeaf (i + head t ft (finite_tf t ft)).
Proof.
  intros t ft bt i.
  rewrite (sk_decompose (sk _ _ _)); simpl.
  generalize (bleaf_tail t ft (finite_tf _ ft) bt). (* introduce the key fact *)
  generalize (tail t ft (finite_tf t ft)). (* less noisy *)
  intros [ s [ nabs fs ]]. (* the only way I could find to make things work! *)
  simpl. intro fact. subst.
  auto.
Qed.

Lemma sk_bleaf:
  forall t, forall ft: finite t, forall i j, sk i t ft = SkLeaf j -> bleaf t.
Proof.
  intros t ft i j.
  rewrite (sk_decompose (sk _ _ _)); simpl.
  intro bt.
  apply (tail_bleaf _ ft (finite_tf t ft)).
  generalize bt. clear bt.
  generalize (tail t ft (finite_tf t ft)). (* less noisy *)
  intros [ s [ nabs fs ]].
  simpl.
  dependent destruction s.
  auto.
  intro. discriminate.
  dependent destruction nabs.
Qed.

(* If [t] satisfies [bleaf] and [u] is above [t], then [u] satisfies [bleaf]
   as well. That is, [bleaf] is upper closed. *)

Lemma bleaf_sksub_bleaf:
  forall i j t u ft fu, bleaf t -> sksub i (sk j t ft) i (sk j u fu) -> bleaf u.
Proof.
  intros ? ? ? ? ? ? bt tu.
  rewrite bleaf_sk in tu; auto.
  dependent destruction tu.
  eapply sk_bleaf. symmetry. eassumption.
Qed.

Lemma bleaf_sub_bleaf:
  forall t u, bleaf t -> sub t u -> bleaf u.
Proof.
  intros ? ? ? stu.
  dependent destruction stu.
  eauto using bleaf_sksub_bleaf. 
Qed.

(* Check that the subtyping relation is non-trivial. We prove this as a
   corollary of the above result, although a direct proof would have been
   easy, since our witnesses are finite types. *)

Lemma sub_nontrivial:
  ~(forall t u, finite t -> finite u -> sub t u).
Proof.
  intro h.
  set (t := TyLeaf).
  set (u := TyArrow TyLeaf TyLeaf).
  assert (finite t) as ft. auto.
  assert (finite u) as fu. unfold u. auto.
  generalize (h t u ft fu); intro tu.
  assert (bleaf u) as bu.
    eapply bleaf_sub_bleaf.
    constructor.
    assumption.
  subst u.
  dependent destruction bu.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Prepending a number of bullets in front of a type. *)

Fixpoint bullets (m : nat) (t : ty) : ty :=
  match m with
  | 0 => t
  | S m => bullets m (TyBullet t)
  end.

Lemma reverse_bullets:
  forall m t, bullets m (TyBullet t) = TyBullet (bullets m t).
Proof.
  induction m; intro; simpl.
  auto.
  rewrite IHm. auto.
Qed.

Lemma bullets_injective:
  forall k t u, bullets k t = bullets k u -> t = u.
Proof.
  induction k; simpl; intros.
  eauto.
  repeat (rewrite reverse_bullets in H).
  apply IHk. congruence.
Qed.

Lemma finite_hereditary_bullets:
  forall k t, finite (bullets k t) -> finite t.
Proof.
  induction k; intros.
  auto.
  apply (finite_hereditary_bullet (TyBullet t)); eauto.
Defined.

Lemma finite_bullets:
  forall k t, finite t -> finite (bullets k t).
Proof.
  induction k; intros; simpl; eauto.
Qed.

Lemma finite_bullets_inversion:
  forall k t, finite (bullets k t) -> finite t.
Proof.
  induction k; simpl; intros.
  eauto.
  rewrite reverse_bullets in H.
  apply IHk.
  eapply finite_hereditary_bullet. eassumption. reflexivity.
Qed.

Hint Resolve finite_bullets.

Lemma sk_bullets:
  forall j t i ft,
  sk i (bullets j t) ft = sk (i + j) t (finite_hereditary_bullets j t ft).
Proof.
  induction j; intros; simpl.
  f_equal. omega.
  rewrite (IHj (TyBullet t)).
  rewrite sk_bullet.
  rewrite arith_plus_successor.
  myEqualUpToPI.
Qed.

Lemma sub_bullets:
  forall k u v, sub u v -> sub (bullets k u) (bullets k v).
Proof.
  induction k; simpl; eauto using sub_bullet.
Qed.

Lemma sub_bullets_intro:
  forall k u v, sub u v -> sub u (bullets k v).
Proof.
  induction k; simpl; eauto using sub_bullet_intro.
Qed.

Lemma sub_arrow_bullets_down:
  forall k u v,
  finite u -> finite v ->
  sub (bullets k (TyArrow u v)) (TyArrow (bullets k u) (bullets k v)).
Proof.
  induction k; simpl; intros.
  eauto.
  eapply sub_transitive.
  eapply sub_bullets.
  eapply sub_arrow_bullet_down; eauto.
  eapply sub_transitive.
  eapply IHk; eauto.
  apply sub_reflexive.
  eauto.
Qed.

Lemma sub_arrow_bullets_up:
  forall k u v,
  finite u -> finite v ->
  sub (TyArrow (bullets k u) (bullets k v)) (bullets k (TyArrow u v)).
Proof.
  induction k; simpl; intros.
  auto.
  eapply sub_transitive.
  eapply IHk; eauto.
  apply sub_bullets.
  apply sub_arrow_bullet_up; eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Subtyping inversion lemmas. *)

Lemma arrow_sub_arrow:
  forall t1 t2 u1 u2,
  sub (TyArrow t1 t2) (TyArrow u1 u2) ->
  exists k,
  sub u1 (bullets k t1) /\ sub (bullets k t2) u2.
Proof.
  intros ? ? ? ? saa.
  dependent destruction saa.
  repeat (rewrite sk_arrow in H).
  dependent destruction H.
  exists k.
  split.
  assert (finite u1) as fu1. eauto.
  assert (finite (bullets k t1)) as fbt1. eauto.
  apply (Sub _ _ fu1 fbt1).
  rewrite sk_bullets. eapply sksub_sk_plus. eapply sksub_pi. eassumption.
  assert (finite (bullets k t2)) as fbt2. eauto.
  assert (finite u2) as fu2. eauto.
  apply (Sub _ _ fbt2 fu2).
  rewrite sk_bullets. eapply sksub_sk_plus. eapply sksub_pi. eassumption.
Qed.

Lemma bullet_sub_bullet:
  forall t u,
  sub (TyBullet t) (TyBullet u) ->
  sub t u.
Proof.
  intros ? ? h.
  dependent destruction h.
  repeat (rewrite sk_bullet in H).
  eapply sksub_sub.
  eassumption.
Qed.

Lemma bullets_sub_bullets:
  forall k t u,
  sub (bullets k t) (bullets k u) ->
  sub t u.
Proof.
  induction k; simpl; intros; eauto using bullet_sub_bullet.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Every finite type can be split into a head and a tail. *)

Lemma head_tail:
  forall t ft tft,
  t = bullets (head t ft tft) (proj1_sig (tail t ft tft)).
Proof.
  intros. dependent induction tft using tf_prop.
  (* Leaf *)
  rewrite head_leaf. rewrite tail_leaf. reflexivity.
  (* Arrow *)
  rewrite head_arrow. rewrite tail_arrow. reflexivity.
  (* Bullet *)
  rewrite head_bullet. rewrite tail_bullet. simpl. rewrite reverse_bullets. f_equal.
  myPunchPI tft.
  myPunchPI (finite_tf t (finite_hereditary_bullet (TyBullet t) ft t refl)).
 apply IHtft. eauto.
Qed.

Lemma finite_head_tail:
  forall t ft tft,
  finite (bullets (head t ft tft) (proj1_sig (tail t ft tft))).
Proof.
  intros. apply finite_bullets. apply finite_tail.
Qed.

Lemma sig_equality:
  forall P,
  forall tt uu : { t: ty | P t },
  proj1_sig tt = proj1_sig uu ->
  tt = uu.
Proof.
  intros.
  dependent destruction tt. dependent destruction uu. simpl. myPunchPI p0. reflexivity.
Qed.

Lemma head_tail_bis:
  forall t ft tft,
  exist _ t ft =
  exist _ (bullets (head t ft tft) (proj1_sig (tail t ft tft))) (finite_head_tail t ft tft).
Proof.
  intros. apply sig_equality. simpl. apply head_tail.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* In Nakano's presentation, there is an explicit requirement that every cycle
   in a recursive type must cross a bullet. Here, instead, we require every
   path that starts at the root node to eventually reach either a leaf or a
   bullet. Furthermore, we require this property to be hereditary. This
   amounts to requiring every path that starts at any interior node to
   eventually reach a leaf or a bullet. *)

Inductive bb : ty -> Prop :=
| BbLeaf : bb TyLeaf
| BbArrow : forall u v, bb u -> bb v -> bb (TyArrow u v)
| BbBullet : forall t, bb (TyBullet t).

Hint Constructors bb.

Definition wf : ty -> Prop :=
  everywhere bb.

Hint Unfold wf.

(* Basic lemmas. *)

Lemma wf_bb:
  forall t, wf t -> bb t.
Proof.
  intros. apply everywhere_root. auto.
Qed.

Hint Resolve wf_bb.

(* Bullet-boundedness is hereditary. *)

Lemma bb_hereditary_domain:
  forall t, bb t -> forall u v, t = TyArrow u v -> bb u.
Proof.
  introv bbt.
  case bbt.
  intros; discriminate.
  intros u v bbu bbv ? ? h. injection h.
  intros ? f; rewrite <- f; assumption.
  intros; discriminate.
Defined.

Lemma bb_hereditary_codomain:
  forall t, bb t -> forall u v, t = TyArrow u v -> bb v.
Proof.
  introv bbt.
  case bbt.
  intros; discriminate.
  intros u v bbu bbv ? ? h. injection h.
  intros f ?; rewrite <- f; assumption.
  intros; discriminate.
Defined.

(* Well-formedness is hereditary. *)

Lemma wf_hereditary_domain:
  forall t, wf t -> forall u v, t = TyArrow u v -> wf u.
Proof.
  intros. subst. eauto.
Qed.

Lemma wf_hereditary_codomain:
  forall t, wf t -> forall u v, t = TyArrow u v -> wf v.
Proof.
  intros. subst. eauto.
Qed.

Lemma wf_hereditary_bullet:
  forall t, wf t -> forall u, t = TyBullet u -> wf u.
Proof.
  intros. subst. eauto.
Qed.

Lemma wf_hereditary_bullets:
  forall k t, wf (bullets k t) -> wf t.
Proof.
  induction k; simpl; intros; eauto.
  rewrite reverse_bullets in H. eauto.
Qed.

Lemma wf_leaf:
  wf TyLeaf.
Proof.
  eauto.
Qed.

Lemma wf_arrow:
  forall u v, wf u -> wf v -> wf (TyArrow u v).
Proof.
  intros. eauto.
  intros. econstructor; eauto.
Qed.

Lemma wf_bullet:
  forall u, wf u -> wf (TyBullet u).
Proof.
  intros. econstructor; eauto.
Qed.

Lemma wf_bullets:
  forall k u, wf u -> wf (bullets k u).
Proof.
  induction k; intros; simpl; eauto using wf_bullet.
Qed.

Hint Resolve wf_arrow wf_bullet wf_bullets.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* An exercise: defining the type Top. In Nakano's presentation, it is
   recursively defined as \mu X, bullet X. *)

CoFixpoint top : ty :=
  TyBullet top.

(* Top is well-formed. *)

Lemma wf_top:
  wf top.
Proof.
  cofix.
  rewrite (ty_decompose top); simpl.
  econstructor; eauto.
Qed.

(* Top is not tail-finite. As a result, it does not have a skeleton, and
   there is not much else we can say about it. *)

Lemma tf_top:
  ~(tf top).
Proof.
  assert (forall t, tyeq t top -> ~(tf t)).
    intros t ttop tft.
    dependent induction tft;
    rewrite (ty_decompose top) in ttop; simpl;
    dependent destruction ttop;
    eauto.
  eauto.
Qed.

