Require Import Setoid.
Require Import Coq.Arith.Compare_dec.
Require Import Coq.Program.Equality.
Require Import Coq.Program.Wf.
Require Export Coq.Logic.ClassicalChoice.
Require Import Coq.Logic.ChoiceFacts.
Require Import LibTactics.
Require Import ArithMisc.
Set Implicit Arguments.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* A consequence of classical logic. *)

Lemma contraposition:
  forall P Q : Prop,
  (~Q -> ~P) -> P -> Q.
Proof.
  intros ? ? ?. apply NNPP. intuition eauto.
Qed.

(* An attempt to prove [forall P, P + ~P]. This does not work, due to
   the fact that eliminating the disjunction [P \/ ~P] is permitted
   only if the goal has sort Prop -- not Set or Type. This does work
   if we introduce a double negation in Prop, but it seems to become
   then rather pointless. *)

Lemma magic:
  forall P : Prop,
  forall A : Prop,
  ((P + ~P) -> A) -> A.
Proof.
  intros P A k.
  generalize (@classic P). intros [ h | h ].
    apply k. left. auto.
    apply k. right. auto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* A theory of infinite sequences. *)

Section Sequence.

  Variable A : Type.

  CoInductive sequence : Type :=
  | SeqCons: sequence -> A -> sequence.

  Definition seq_decomp (s : sequence) : sequence :=
    match s with
    | SeqCons s a => SeqCons s a
    end.

  Lemma seq_decompose:
    forall s, s = seq_decomp s.
  Proof.
    intro s; case s; eauto.
  Qed.

  (* Properties of sequences. *)

  Definition empty :=
    fun s : sequence => False.

  Definition neg (P : sequence -> Prop) : sequence -> Prop :=
    fun s => ~(P s).

  Definition union P Q :=
    fun s : sequence => P s \/ Q s.

  Definition intersection P Q :=
    fun s : sequence => P s /\ Q s.

  Hint Unfold empty neg union intersection subset.

  (* Rewriting over predicates. *)

  Lemma subset_reflexive:
    forall P : sequence -> Prop, subset P P.
  Proof.
    intuition eauto.
  Qed.

  Lemma subset_transitive:
    forall P Q R : sequence -> Prop, subset P Q -> subset Q R -> subset P R.
  Proof.
    intuition eauto.
  Qed.

  Definition ss := (@subset sequence).

  Lemma NN:
    forall P,
    subset (neg (neg P)) P.
  Proof.
    intros P s. unfold neg. apply NNPP.
  Qed.

  Lemma NN_back:
    forall P,
    subset P (neg (neg P)).
  Proof.
    unfold neg. intuition eauto.
  Qed.

  Lemma contraposition_subset:
    forall P Q,
    subset (neg Q) (neg P) ->
    subset P Q.
  Proof.
    unfold subset, neg.
    intros ? ? ? ?.
    apply contraposition.
    auto.
  Qed.

  (* A sequence satisfies [now P] if and only if its first element satisfies
     [P]. *)

  Definition now (P : A -> Prop) :=
    fun s => match s with SeqCons _ a => P a end.

  (* A sequence satisfies [later P] if and only if its tail satisfies [P]. *)

  Definition later (P : sequence -> Prop) :=
    fun s => match s with SeqCons s _ => P s end.

  (* A sequence satisfies [rel R] if and only if its first two elements are
     related by [R]. *)

  Definition rel (R : A -> A -> Prop) : sequence -> Prop :=
    fun s => match s with SeqCons (SeqCons _ a) b => R a b end.

  (* A sequence satisfies [always P] if and only if every suffix of it satisfies [P]. *)

  CoInductive always (P : sequence -> Prop) : sequence -> Prop :=
  | Always: forall s a, always P s -> P (SeqCons s a) -> always P (SeqCons s a).

  (* A sequence satisfies [eventually P] if and only if some suffix of it satisfies [P]. *)

  Inductive eventually (P : sequence -> Prop) : sequence -> Prop :=
  | EventuallyNow: forall s, P s -> eventually P s
  | EventuallyLater: forall s a, eventually P s -> eventually P (SeqCons s a).

  (* Basic properties. *)

  Lemma A_cons_head:
    forall P s,
    always P s ->
    P s.
  Proof.
    introv h. dependent destruction h. auto.
  Qed.

  Lemma A_cons_tail:
    forall P s a,
    always P (SeqCons s a) ->
    always P s.
  Proof.
    introv h. dependent destruction h. auto.
  Qed.

  Lemma A_cons:
    forall P s a,
    always P (SeqCons s a) ->
    P (SeqCons s a) /\ always P s.
  Proof.
    intros. split. eapply A_cons_head. auto. eapply A_cons_tail. eauto.
  Qed.

  Lemma NE_cons_head:
    forall P s,
    ~(eventually P s) ->
    ~(P s).
  Proof.
    introv h. intro.
    apply h. apply EventuallyNow. auto.
  Qed.

  Lemma NE_cons_tail:
    forall P s a,
    ~(eventually P (SeqCons s a)) ->
    ~(eventually P s).
  Proof.
    introv h. intro.
    apply h. apply EventuallyLater. auto.
  Qed.

  Lemma A_covariant:
    forall P Q,
    subset P Q ->
    subset (always P) (always Q).
  Proof.
    unfold subset.
    cofix.
    intros ? ? h s aps.
    dependent destruction s. dependent destruction aps.
    constructor; eauto.
  Qed.

  Lemma E_covariant:
    forall P Q,
    subset P Q ->
    subset (eventually P) (eventually Q).
  Proof.
    introv h. unfold subset. induction 1.
    apply EventuallyNow. auto.
    apply EventuallyLater. auto.
  Qed.

End Sequence.

Hint Resolve A_cons_tail NE_cons_tail.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Now seems a good time to set up a little theory of rewriting over
   subset assertions. (These declarations must lie outside a section.) *)

Hint Unfold empty neg union intersection subset ss rel now later.

Add Parametric Relation (A : Type) : (sequence A -> Prop) (@ss A)
  reflexivity proved by (@subset_reflexive A)
  transitivity proved by (@subset_transitive A)
as subset_setoid.

Add Parametric Morphism (A : Type) : (@neg A) with signature (flip (@ss A)) ==> (@ss A) as neg_morphism.
Proof.
  unfold neg.
  intuition eauto.
Qed.

Add Parametric Morphism (A : Type) : (@union A) with signature (@ss A) ==> (@ss A) ==> (@ss A) as union_morphism.
Proof.
  unfold union, ss, subset.
  intuition eauto.
Qed.

Add Parametric Morphism (A : Type) : (@intersection A) with signature (@ss A) ==> (@ss A) ==> (@ss A)
as intersection_morphism.
Proof.
  unfold intersection, ss, subset.
  intuition eauto.
Qed.

Add Parametric Morphism (A : Type) : (@ss A) with signature (flip (@ss A)) ==> (@ss A) ==> impl as subset_morphism.
Proof.
  unfold ss, impl.
  intuition eauto.
Qed. (* TEMPORARY useless? check *)

Add Parametric Morphism (A : Type) : (@always A) with signature (@ss A) ==> (@ss A) as always_morphism.
Proof.
  apply A_covariant.
Qed.

Add Parametric Morphism (A : Type) : (@eventually A) with signature (@ss A) ==> (@ss A) as eventually_morphism.
Proof.
  apply E_covariant.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* A little theory of temporal logic over sequences. *)

Lemma to_ss:
  forall A (P Q : sequence A -> Prop) s,
  P s ->
  ss P Q ->
  Q s.
Proof.
  introv ? h. apply h. auto.
Qed.

Lemma A_and_A:
  forall A P Q (s : sequence A),
  always P s -> always Q s -> always (intersection P Q) s.
Proof.
  cofix.
  intros ? ? ? [ s a ] p q.
  dependent destruction p. dependent destruction q.
  constructor; auto.
Qed.

Lemma IAA:
  forall A P Q,
  @ss A (intersection (always P) (always Q)) (always (intersection P Q)).
Proof.
  intros ? ? ? s.
  unfold intersection at 1.
  intros [ ? ? ].
  eapply A_and_A; eauto.
Qed.

Lemma EA_and_EA:
  forall A P Q (s : sequence A),
  eventually (always P) s -> eventually (always Q) s -> eventually (always (intersection P Q)) s.
Proof.
  induction 1 as [ ? eps | ? ? eaps ihP ].
  induction 1 as [ ? eqs | ? ? eaqs ihQ ].
  (* case 1 *)
  apply EventuallyNow. apply A_and_A; auto.
  (* case 2 *)
  apply EventuallyLater. apply ihQ. dependent destruction eps. auto.
  (* case 3 *)
  intro eaqsa. apply EventuallyLater. apply ihP.
  dependent destruction eaqsa.
  (* subcase a *)
  apply EventuallyNow. eauto.
  (* subcase b *)
  eauto.
Qed.

Lemma IEAEA:
  forall A P Q,
  @ss A (intersection (eventually (always P)) (eventually (always Q))) (eventually (always (intersection P Q))).
Proof.
  intros ? ? ? s.
  unfold intersection at 1.
  intros [ ? ? ].
  eapply EA_and_EA; eauto.
Qed.

Lemma A_false:
  forall A,
  subset (always (@empty A)) (@empty A).
Proof.
  intros ? s h. dependent destruction h. auto.
Qed.

Lemma A_false_variant:
  forall A P s,
  always P s ->
  subset P (@empty A) ->
  empty s.
Proof.
  intros ? ? ? ap pa.
  eapply to_ss. eexact ap.
  rewrite pa.
  apply A_false.
Qed.

Lemma E_false:
  forall A,
  subset (eventually (@empty A)) (@empty A).
Proof.
  intro. intro. induction 1; auto.
Qed.

Lemma E_false_variant:
  forall A P s,
  eventually P s ->
  subset P (@empty A) ->
  empty s.
Proof.
  intros ? ? ? ep pa.
  eapply to_ss. eexact ep.
  rewrite pa.
  apply E_false.
Qed.

Lemma I_commutative:
  forall A P Q,
  @ss A
    (intersection P Q)
    (intersection Q P).
Proof.
  intros. intro. unfold intersection. intuition eauto.
Qed.

Lemma EA_false:
  forall A,
  subset (eventually (always (@empty A))) (@empty A).
Proof.
  intro. rewrite A_false. apply E_false.
Qed.

Lemma AA:
  forall A P,
  @ss A
    (always P) (always (always P)).
Proof.
  intros ? ?. cofix.
  intros [ s a ] ?.
  constructor.
  dependent destruction H. apply AA. auto.
  auto.
Qed.

Lemma EE:
  forall A P,
  @ss A
    (eventually (eventually P)) (eventually P).
Proof.
  intros ? ? s. induction 1. auto. apply EventuallyLater. auto.
Qed.

Lemma A_and_E:
  forall A P Q (s : sequence A),
  always P s ->
  eventually Q s ->
  eventually (intersection P Q) s.
Proof.
  induction 2.
  apply EventuallyNow. dependent destruction H. eauto.
  apply EventuallyLater. eauto.
Qed.

Lemma A_and_E_variant:
  forall A P Q (s : sequence A),
  always P s ->
  eventually Q s ->
  eventually (intersection (always P) Q) s.
Proof.
  intros.
  eapply A_and_E.
  apply AA. auto.
  auto.
Qed.

Lemma IAE:
  forall A P Q,
  @ss A
    (intersection (always Q) (eventually P))
    (eventually (intersection (always Q) P)).
Proof.
  intros ? ? ? s [ ? ? ].
  eapply A_and_E_variant; eauto.
Qed.

Lemma AN_and_E:
  forall A P (s : sequence A),
  always (neg P) s ->
  eventually P s ->
  False.
Proof.
  unfold neg.
  induction 2.
  dependent destruction H. auto.
  eauto.
Qed.

Lemma A_and_EN:
  forall A P (s : sequence A),
  always P s ->
  eventually (neg P) s ->
  False.
Proof.
  unfold neg.
  induction 2.
  dependent destruction H. auto.
  eauto.
Qed.

Lemma NE:
  forall A P,
  @ss A (neg (eventually P)) (always (neg P)).
Proof.
  unfold ss, subset.
  cofix.
  intros ? P [ s a ] h.
  constructor. eauto.
  apply NE_cons_head. auto.
Qed.

Lemma NEN:
  forall A P,
  @ss A (neg (eventually (neg P))) (always P).
Proof.
  intros. rewrite NE. rewrite NN. auto.
Qed.

Lemma NA:
  forall A P,
  @ss A (neg (always P)) (eventually (neg P)).
Proof.
  intros.
  apply contraposition_subset.
  rewrite NEN. apply NN_back.
Qed.

Lemma NAN:
  forall A P,
  @ss A (neg (always (neg P))) (eventually P).
Proof.
  intros.
  rewrite NA. rewrite NN. auto.
Qed.

Lemma NEN_right:
  forall A P,
  @ss A (always P) (neg (eventually (neg P))).
Proof.
  intros ? P s a e.
  eapply A_and_EN; eauto.
Qed.

Lemma NA_right:
  forall A P,
  @ss A (eventually (neg P)) (neg (always P)).
Proof.
  intros.
  apply contraposition_subset.
  rewrite NN.
  apply NEN_right.
Qed.

Lemma NU:
  forall A P Q,
  @ss A (neg (union P Q)) (intersection (neg P) (neg Q)).
Proof.
  intros ? ? ? s.
  unfold neg, union, intersection.
  eauto using not_or_and.
Qed.

Lemma NU_right:
  forall A P Q,
  @ss A (intersection (neg P) (neg Q)) (neg (union P Q)).
Proof.
  intros ? ? ? s.
  unfold neg, union, intersection.
  eauto using and_not_or.
Qed.

Lemma A_intro:
  forall A P,
  @ss A (always P) P.
Proof.
  intros ? P s h.
  dependent destruction h.
  auto.
Qed.

Lemma A_intro_contra:
  forall A P (s : sequence A),
  ~ P s ->
  ~ always P s.
Proof.
  intros ? ? ? nps aps.
  apply nps. apply A_cons_head. auto.
Qed.

Lemma E_intro:
  forall A P,
  @ss A P (eventually P).
Proof.
  intros ? ? ? ?. apply EventuallyNow. auto.
Qed.

Lemma EL:
  forall A P,
  @ss A
    (eventually (later P)) (eventually P).
Proof.
  intros ? ? s. induction 1.
    destruct s as [ s a ]. apply EventuallyLater. apply E_intro. auto.
    apply EventuallyLater. auto.
Qed.

Lemma A_union:
  forall A P Q,
    @ss A
      (always (union P Q))
      (union (always (eventually P)) (always (eventually Q))).
Proof.
  intros.
  apply contraposition_subset.
  rewrite NU.
  rewrite NA. rewrite NE. rewrite NA. rewrite NE.
  rewrite IEAEA.
  rewrite <- NA_right.
  rewrite <- NU_right.
  rewrite A_intro.
  auto.
Qed.

Lemma A_union_explicit:
  forall A P Q (s : sequence A),
  always (union P Q) s ->
  always (eventually P) s \/ always (eventually Q) s.
Proof.
  intros ? ? ? ? h.
  generalize (A_union h).
  auto.
Qed.

Lemma subset_neg:
  forall A P Q,
  subset (intersection P Q) (@empty A) ->
  subset P (neg Q).
Proof.
  intros ? ? ? h s ps qs.
  unfold subset, empty in h.
  apply (h s). eauto.
Qed.

Lemma subset_neg_neg:
  forall A P Q,
  subset (intersection P (neg Q)) (@empty A) ->
  subset P Q.
Proof.
  intros ? ? ? h.
  rewrite <- (@NN _ Q).
  apply subset_neg. auto.
Qed.

Lemma subset_reinforce:
  forall A P Q,
  subset P Q ->
  @ss A P (intersection P Q).
Proof.
  intros ? ? s ?. split; auto.
Qed.

Lemma AEA_reinforce:
  forall A P Q,
  @ss A (always P) (eventually (always Q)) ->
  @ss A (always P) (eventually (always (intersection P Q))).
Proof.
  introv h.
  rewrite (subset_reinforce h).
  rewrite IAE.
  apply E_covariant.
  rewrite IAA.
  apply A_covariant.
  auto.
Qed.

Lemma AEA_reinforce_variant:
  forall A P Q R,
  @ss A (always P) (eventually (always Q)) ->
  @ss A (intersection P Q) R ->
  @ss A (always P) (eventually (always R)).
Proof.
  introv h pqr.
  rewrite <- pqr.
  apply AEA_reinforce.
  auto.
Qed.

(* If one cannot have forever [P] without [Q], then [always P] implies
   [often Q]. *)

Lemma always_often:
  forall A P Q,
  subset (always (intersection P (neg Q))) (@empty A) ->
  subset (always P) (always (eventually Q)).
Proof.
  intros ? ? ? pnq.
  rewrite <- (@NN _ (always (eventually Q))).
  rewrite NA. rewrite (@NE _ Q). apply subset_neg.
  rewrite (@E_intro _ (always P)).
  rewrite IEAEA.
  rewrite pnq.
  apply E_false.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* A sequence satisfies [increasingly P] iff it satisfies [eventually (always
   (P k))] for arbitrarily large values of [k]. *)

Definition increasingly (A : Type) (P : nat -> sequence A -> Prop) (s : sequence A) : Prop :=
  forall k,
  eventually (always (P k)) s.

Lemma step_increasingly_preliminary:
  forall A (P : nat -> sequence A -> Prop) (Q : sequence A -> Prop),
  (forall k s, P k s -> always Q s -> eventually (P (k + 1)) s) ->
  forall j k,
  subset
    (intersection (always Q) (P k))
    (eventually (intersection (always Q) (P (k + j)))).
Proof.
  induction j; intros.
  rewrite arith_plus_zero. intros ? [ ? ? ]. apply EventuallyNow. auto.
  replace (k + S j) with ((k + 1) + j); try omega.
  rewrite <- EE.
  rewrite <- (IHj (k + 1)).
  intros s [ ? ? ].
  apply A_and_E_variant. auto.
  eauto.
Qed.

Lemma step_increasingly:
  forall A (P : nat -> sequence A -> Prop) (Q : sequence A -> Prop),
  (forall m n, m <= n -> subset (P n) (P m)) ->                      (* monotonicity *)
  (forall k s, P k s -> always Q s -> always (P k) s) ->             (* preservation *)
  (forall k s, P k s -> always Q s -> eventually (P (k + 1)) s) ->   (* step *)
  forall k,
  subset
    (intersection (always Q) (P k))
    (increasingly P).
Proof.
  intros ? ? ? mono preservation step ?.
  intros s [ ? ? ] n.
  case (le_gt_dec n k); intro.
  (* n <= k *)
  apply EventuallyNow. apply preservation. eapply mono; eassumption. eassumption.
  (* n > k *)
  generalize (step_increasingly_preliminary P step (n - k) k (conj H H0)); intro ep.
  replace (k + (n - k)) with n in ep; try omega.
  eapply to_ss. eexact ep.
  apply E_covariant.
  intros ? [ ? ? ]. eapply preservation; eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* TEMPORARY nettoyer la suite *)

Lemma totality:
  forall A : Type,
  forall P : A -> Prop,
  forall R : A -> A -> Prop,
  (forall x, P x -> exists y, P y /\ R x y) ->
  (forall x : sig P, exists y : sig P, R (proj1_sig x) (proj1_sig y)).
Proof.
  intros ? ? ? h [ x px ].
  generalize (h x px); intros [ y [ py rxy ]].
  exists (exist _ y py).
  auto.
Qed.

CoFixpoint iterate (A : Type) (P : A -> Prop) (f : sig P -> sig P) (x : sig P) : sequence A :=
  SeqCons
    (iterate f (f x))
    (proj1_sig x).

Lemma iterate_related:
  forall A : Type,
  forall P : A -> Prop,
  forall R : A -> A -> Prop,
  forall f : sig P -> sig P,
  (forall x, R (proj1_sig (f x)) (proj1_sig x)) ->
  forall x,
  always (rel R) (iterate f x).
Proof.
  introv h.
  cofix.
  intro.
  rewrite (seq_decompose (iterate f x)). simpl.
  constructor. eauto.
  unfold rel.
  rewrite (seq_decompose (iterate f (f x))). simpl.
  apply h.
Qed.

Lemma iterate_now:
  forall A : Type,
  forall P : A -> Prop,
  forall f : sig P -> sig P,
  forall x,
  always (now P) (iterate f x).
Proof.
  cofix.
  intros.
  rewrite (seq_decompose (iterate f x)); simpl.
  constructor. eauto.
  simpl.
  exact (proj2_sig x).
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

Definition tl A (s : sequence A) :=
  match s with
  | SeqCons s _ => s
  end.

CoFixpoint map A B (f : A -> B) (xs : sequence A) : sequence B :=
  match xs with
  | SeqCons xs x =>
      SeqCons (map f xs) (f x)
  end.

CoFixpoint zip (A : Type) (f : A -> A -> A) (s1 s2 : sequence A) : sequence A :=
  match s1, s2 with
  | SeqCons s1 a1, SeqCons s2 a2 =>
      SeqCons (zip f s1 s2) (f a1 a2)
  end.

Lemma zip_monotonic:
  forall A (f : A -> A -> A) (R1 R2 R : A -> A -> Prop),
  (forall y1 y2 x1 x2, R1 y1 x1 -> R2 y2 x2 -> R(f y1 y2) (f x1 x2)) ->
  forall s1 s2,
  always (rel R1) s1 ->
  always (rel R2) s2 ->
  always (rel R) (zip f s1 s2).
Proof.
  introv m. cofix. intros [[ s1 y1 ] x1 ] [[ s2 y2 ] x2 ] h1 h2.
  rewrite (seq_decompose (zip f (SeqCons (SeqCons s1 y1) x1) (SeqCons (SeqCons s2 y2) x2))). simpl. constructor.
  eauto.
  rewrite (seq_decompose (zip f (SeqCons s1 y1) (SeqCons s2 y2))). simpl. eauto.
  generalize (A_cons_head h1). generalize (A_cons_head h2). eauto.
Qed.

CoFixpoint orbit (A : Type) (f : A -> A) (x : A) : sequence A :=
  SeqCons (orbit f (f x)) x.

(* ----------------------------------------------------------------------------------------------------------------- *)

CoInductive bisimilar A : sequence A -> sequence A -> Prop :=
  | BiSimilar : forall s1 s2 x1 x2, bisimilar s1 s2 -> x1 = x2 -> bisimilar (SeqCons s1 x1) (SeqCons s2 x2).

Hint Constructors bisimilar.

Lemma bisimilar_symmetric:
  forall A (s1 s2 : sequence A),
  bisimilar s1 s2 ->
  bisimilar s2 s1.
Proof.
  cofix. intros ? [ s1 x1 ] [ s2 x2 ] h. dependent destruction h. constructor; eauto.
Qed.

Hint Immediate bisimilar_symmetric.

Definition compatible A (P : sequence A -> Prop) :=
  forall s1 s2,
  P s1 ->
  bisimilar s1 s2 ->
  P s2.

Implicit Arguments compatible [ A ].

Lemma now_compatible:
  forall A (P : A -> Prop),
  compatible (now P).
Proof.
  intros ? ? [ s1 x1 ] [ s2 x2 ]. unfold now. intros ? h. dependent destruction h. auto.
Qed.

Lemma always_compatible:
  forall A (P : sequence A -> Prop),
  compatible P ->
  compatible (always P).
Proof.
  intros ? ? ?. unfold compatible. cofix. intros [ s1 x1 ] [ s2 x2 ] a b.
  dependent destruction a. dependent destruction b. constructor; eauto.
Qed.

Lemma eventually_compatible:
  forall A (P : sequence A -> Prop),
  compatible P ->
  compatible (eventually P).
Proof.
  intros ? ? ?. unfold compatible. intros s1 s2 h1. generalize s2. clear s2.
  induction h1; intros ? h2.
  apply EventuallyNow. eauto.
  dependent destruction h2. apply EventuallyLater. eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

Section Chains.

  Variable A : Type.

  Variable R : A -> A -> Prop.

  (* A set [P] forms a (possibly empty) descending chain if and only if every
     element of [P] admits a predecessor in [P]. *)

  Definition chain (P : A -> Prop) : Prop :=
    forall b, P b -> exists a, P a /\ R a b.

  (* An element is inaccessible if and only if it is not accessible. *)

  Definition inaccessible x :=
    ~(Acc R x).

  (* Inaccessibility forms a chain. *)

  Lemma chain_inaccessible:
    chain inaccessible.
  Proof.
    intros b h.
    (* It is not the case that every predecessor of [b] is accessible. *)
    assert (~(forall a, R a b -> Acc R a)) as e.
      intro. apply h. constructor. auto.
    (* Hence, some predecessor must be inaccessible. *)
    generalize (not_all_ex_not _ _ e); intros [ a f ].
    exists a.
    generalize (imply_to_and _ _ f); intros [ ? ? ].
    auto.
  Qed.

  (* If [R] is well-founded, then [inaccessible] is empty. (This is intuitionistic.) *)

  Lemma wf_inaccessible_empty:
    well_founded R -> ~(exists x, inaccessible x).
  Proof.
    intro. apply all_not_not_ex. auto.
  Qed.

  (* The converse also holds. (This is classical.) *)

  Lemma inaccessible_empty_wf:
    ~(exists x, inaccessible x) -> well_founded R.
  Proof.
    intro. unfold well_founded. apply not_ex_not_all. auto.
  Qed.

  (* An infinite descending path. *)

  CoInductive path (x : A) : Prop :=
  | Cons: forall y, path y -> R y x -> path x.

  (* Every inaccessible element forms the head of an infinite descending path. *)

  Lemma inaccessible_path:
    forall x,
    inaccessible x ->
    path x.
  Proof.
    cofix.
    intros ? h.
    generalize (chain_inaccessible h); intros [ y [ ? ? ]].
    econstructor; eauto.
  Qed.

  (* The converse is true. *)

  Lemma path_inaccessible:
    forall x,
    path x ->
    inaccessible x.
  Proof.
    assert (forall x, Acc R x -> path x -> False) as fact.
      induction 1.
      intro h.
      dependent destruction h.
      eauto.
    intro.
    apply contraposition.
    intro. intro.
    eapply fact.
    apply NNPP.
    eassumption.
    auto.
  Qed.

  Lemma inaccessible_sequence:
    forall a,
    inaccessible a ->
    exists s,
    always (rel R) (SeqCons s a).
  Proof.
    introv pa.
    (* Apply the axiom of choice to the reverse of [R], restricted to the inaccessible elements. *)
    set (P := inaccessible).
    set (Rr := fun a b => R b a).
    generalize (
      @choice (sig P) (sig P)
      (fun (x : sig P) (y : sig P) => R (proj1_sig y) (proj1_sig x))
      (totality Rr chain_inaccessible)
    ); intros [ next pnext ].
    (* This yields a function [next]. Iterate this function to obtain an
       infinite descending sequence. *)
    exists (iterate next (next (exist _ a pa))).
    (* There remains to show that the elements in this sequence are related
       by [R]. *)
    replace
      (SeqCons (iterate next (next (exist _ a pa))) a)
    with
      (iterate next (exist _ a pa)).
    apply iterate_related.
    auto.
    rewrite (seq_decompose (iterate next (exist _ a pa))). reflexivity.
  Qed.

  Lemma if_sequence:
    ~(well_founded R) ->
    exists s,
    always (rel R) s.
  Proof.
    intro h.
    generalize (not_all_ex_not _ _ h); intros [ a ia ].
    generalize (inaccessible_sequence ia); intros [ s ? ].
    exists (SeqCons s a). auto.
  Qed.

  Lemma sequence_inaccessible:
    forall s a,
    always (rel R) (SeqCons s a) ->
    inaccessible a.
  Proof.
    assert (forall a, Acc R a -> forall s, always (rel R) (SeqCons s a) -> False).
      induction 1 as [ a _ ih ].
      intro s. dependent destruction s.
      intro h.
      assert (R a0 a).
        dependent destruction h. auto.
      eauto using A_cons_tail.
    intros.
    unfold inaccessible.
    eauto.
  Qed.

  Lemma sequence_if:
    forall s,
    always (rel R) s ->
    ~(well_founded R).
  Proof.
    intro.
    dependent destruction s.
    intro.
    forwards: sequence_inaccessible. eassumption.
    apply ex_not_not_all. eauto.
  Qed.

End Chains.

